from typing import TYPE_CHECKING, Any, Tuple, Union
import numpy

from amulet.api.wrapper import Interface
from .schematic import SchematicChunk
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionBox
from amulet.world_interface.chunk import translators
from amulet.api.data_types import AnyNDArray

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator
    from PyMCTranslate import TranslationManager


class SchematicInterface(Interface):
    _entity_id_type = None
    _entity_coord_type = None
    _block_entity_id_type = None
    _block_entity_coord_type = None

    def is_valid(self, key: Tuple) -> bool:
        return True

    def decode(
        self, cx: int, cz: int, section: SchematicChunk
    ) -> Tuple["Chunk", AnyNDArray]:
        chunk = Chunk(cx, cz)
        block_palette, blocks = numpy.unique(
            (section.blocks << 4) + (section.data & 0xF), return_inverse=True
        )
        blocks = blocks.reshape(section.blocks.shape)
        palette = numpy.empty(len(block_palette) + 1, dtype=numpy.object)
        palette[0] = (0, 0)
        for index, block_num in enumerate(block_palette):
            palette[index + 1] = (block_num >> 4, block_num & 0xF)

        box = section.selection.create_moved_box((cx * 16, 0, cz * 16), subtract=True)
        chunk.blocks[box.slice] = blocks + 1
        for b in section.block_entities:
            b = self._decode_block_entity(
                b, self._block_entity_id_type, self._block_entity_coord_type
            )
            if b is not None:
                chunk.block_entities.insert(b)
        for b in section.entities:
            b = self._decode_entity(
                b, self._block_entity_id_type, self._block_entity_coord_type
            )
            if b is not None:
                chunk.entities.append(b)

        return chunk, palette

    def encode(
        self,
        chunk: "Chunk",
        palette: AnyNDArray,
        max_world_version: Tuple[str, Union[int, Tuple[int, int, int]]],
        box: SelectionBox = None,
    ) -> SchematicChunk:
        entities = []
        for e in chunk.entities:
            if e.location in box:
                entities.append(
                    self._encode_entity(
                        e, self._entity_id_type, self._entity_coord_type
                    ).value
                )
        block_entities = []
        for e in chunk.block_entities:
            if e.location in box:
                block_entities.append(
                    self._encode_block_entity(
                        e, self._block_entity_id_type, self._block_entity_coord_type
                    ).value
                )

        slices = box.create_moved_box(
            (chunk.cx * 16, 0, chunk.cz * 16), subtract=True
        ).slice
        blocks_merged = palette[chunk.blocks[slices]]

        return SchematicChunk(
            box,
            blocks_merged[:, :, :, 0].astype(numpy.uint16),
            blocks_merged[:, :, :, 1].astype(numpy.uint8),
            block_entities,
            entities,
        )

    def get_translator(
        self,
        max_world_version: Tuple[str, Tuple[int, int, int]],
        data: Any = None,
        translation_manager: "TranslationManager" = None,
    ) -> Tuple["Translator", Union[int, Tuple[int, int, int]]]:
        platform, version_number = max_world_version
        if platform == "java":
            version = translation_manager.get_version(platform, version_number)
            version_number = version.data_version
        return translators.loader.get((platform, version_number)), version_number


class JavaSchematicInterface(SchematicInterface):
    _entity_id_type = "namespace-str-id"
    _entity_coord_type = "Pos-list-double"
    _block_entity_id_type = "namespace-str-id"
    _block_entity_coord_type = "xyz-int"


class BedrockSchematicInerface(SchematicInterface):
    _entity_id_type = "namespace-str-identifier"
    _entity_coord_type = "Pos-list-float"
    _block_entity_id_type = "str-id"
    _block_entity_coord_type = "xyz-int"



from __future__ import annotations

from typing import Tuple, Any, Union, Generator, Dict, List, TYPE_CHECKING
import copy
import numpy

import PyMCTranslate

from amulet import log
from amulet.api.registry import BlockManager
from amulet.world_interface.chunk import interfaces
from amulet.api.errors import (
    ChunkLoadError,
    ChunkDoesNotExist,
    ObjectReadWriteError,
)
from amulet.api.data_types import (
    AnyNDArray,
    VersionNumberAny,
    ChunkCoordinates,
)

if TYPE_CHECKING:
    from amulet.api.wrapper import Interface
    from amulet.api.chunk import Chunk
    from amulet.api.wrapper.chunk.translator import Translator


class BaseFormatWrapper:
    """
    The Format class is a class that sits between the serialised world or structure data and the program using amulet-core.
    The Format class is used to access chunks from the serialised source in the universal format and write them back again.
    The BaseFormat class holds the common methods shared by the sub-classes.
    """

    def __init__(self):
        self._translation_manager = None
        self._version = None
        self._changed: bool = False

    @property
    def readable(self) -> bool:
        """Can this object have data read from it."""
        return True

    @property
    def writeable(self) -> bool:
        """Can this object have data written to it."""
        return True

    @property
    def translation_manager(self) -> PyMCTranslate.TranslationManager:
        """The translation manager attached to the world"""
        if self._translation_manager is None:
            self._translation_manager = PyMCTranslate.new_translation_manager()
        return self._translation_manager

    @translation_manager.setter
    def translation_manager(self, value: PyMCTranslate.TranslationManager):
        self._translation_manager = value

    @staticmethod
    def is_valid(path: str) -> bool:
        """
        Returns whether this format is able to load the given object.

        :param path: The path of the object to load.
        :return: True if the world can be loaded by this format, False otherwise.
        """
        raise NotImplementedError()

    @property
    def platform(self) -> str:
        """Platform string ("bedrock" / "java" / ...)"""
        raise NotImplementedError

    @property
    def version(self) -> Union[int, Tuple[int, ...]]:
        """Platform string ("bedrock" / "java" / ...)"""
        if self._version is None:
            self._version = self._get_version()
        return self._version

    def _get_version(self) -> Union[int, Tuple[int, ...]]:
        raise NotImplementedError

    @property
    def max_world_version(self) -> Tuple[str, Union[int, Tuple[int, ...]]]:
        """The version the world was last opened in
        This should be greater than or equal to the chunk versions found within"""
        return self.platform, self.version

    @property
    def changed(self) -> bool:
        return self._changed

    def _get_interface(self, max_world_version, raw_chunk_data=None) -> "Interface":
        if raw_chunk_data:
            key = self._get_interface_key(raw_chunk_data)
        else:
            key = max_world_version
        return interfaces.loader.get(key)

    def _get_interface_and_translator(
        self, max_world_version, raw_chunk_data=None
    ) -> Tuple["Interface", "Translator", "VersionNumberAny"]:
        interface = self._get_interface(max_world_version, raw_chunk_data)
        translator, version_identifier = interface.get_translator(
            self.max_world_version, raw_chunk_data
        )
        return interface, translator, version_identifier

    def _get_interface_key(self, raw_chunk_data) -> Any:
        raise NotImplementedError

    def open(self):
        """Open the database for reading and writing"""
        raise NotImplementedError

    @property
    def has_lock(self) -> bool:
        """Verify that the world database can be read and written"""
        raise NotImplementedError

    def _verify_has_lock(self):
        """Ensure that the Format has a lock on the world. Throw WorldAccessException if not"""
        if not self.has_lock:
            raise ObjectReadWriteError(
                "The world has been opened somewhere else or the .open() method was not called"
            )

    def save(self):
        """Save the data back to the disk database"""
        raise NotImplementedError

    def close(self):
        """Close the disk database"""
        raise NotImplementedError

    def unload(self):
        """Unload data stored in the Format class"""
        raise NotImplementedError

    def all_chunk_coords(self, *args) -> Generator[ChunkCoordinates, None, None]:
        """A generator of all chunk coords"""
        raise NotImplementedError

    def load_chunk(self, cx: int, cz: int, *args) -> "Chunk":
        """
        Loads and creates a universal amulet.api.chunk.Chunk object from chunk coordinates.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :return: The chunk at the given coordinates.
        """
        if not self.readable:
            raise ChunkLoadError("This object is not readable")
        try:
            return self._load_chunk(cx, cz, *args)
        except ChunkDoesNotExist as e:
            raise e
        except Exception:
            log.error(f"Error loading chunk {cx} {cz}", exc_info=True)
            raise ChunkLoadError

    def _load_chunk(self, cx: int, cz: int, *args, recurse: bool = True,) -> "Chunk":
        """
        Loads and creates a universal amulet.api.chunk.Chunk object from chunk coordinates.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param recurse: bool: look in boundary chunks if required to fully define data
        :return: The chunk at the given coordinates.
        """

        # Gets an interface (the code that actually reads the chunk data)
        raw_chunk_data = self._get_raw_chunk_data(cx, cz, *args)
        interface, translator, game_version = self._get_interface_and_translator(
            self.max_world_version, raw_chunk_data
        )

        # decode the raw chunk data into the universal format
        chunk, chunk_palette = self._decode(interface, cx, cz, raw_chunk_data)
        chunk_palette: AnyNDArray
        chunk = self._unpack(translator, game_version, chunk, chunk_palette)
        return self._convert_to_load(
            chunk, translator, game_version, *args, recurse=recurse,
        )

    def _decode(
        self, interface: "Interface", cx: int, cz: int, raw_chunk_data: Any
    ) -> Tuple["Chunk", AnyNDArray]:
        return interface.decode(cx, cz, raw_chunk_data)

    def _unpack(
        self,
        translator: "Translator",
        game_version: VersionNumberAny,
        chunk: "Chunk",
        chunk_palette: AnyNDArray,
    ) -> "Chunk":
        return translator.unpack(
            game_version, self.translation_manager, chunk, chunk_palette
        )

    def _convert_to_load(
        self,
        chunk: "Chunk",
        translator: "Translator",
        game_version: VersionNumberAny,
        *args,
        recurse: bool = True,
    ) -> "Chunk":
        # set up a callback that translator can use to get chunk data
        cx, cz = chunk.cx, chunk.cz
        if recurse:
            chunk_cache: Dict[ChunkCoordinates, "Chunk"] = {}

            def get_chunk_callback(x: int, z: int) -> "Chunk":
                cx_, cz_ = cx + x, cz + z
                if (cx_, cz_) not in chunk_cache:
                    chunk_cache[(cx_, cz_)] = self._load_chunk(
                        cx_, cz_, *args, recurse=False
                    )
                return chunk_cache[(cx_, cz_)]

        else:
            get_chunk_callback = None

        # translate the data to universal format
        chunk = translator.to_universal(
            game_version, self.translation_manager, chunk, get_chunk_callback, recurse,
        )

        chunk.changed = False
        return chunk

    def commit_chunk(self, chunk: "Chunk", *args):
        """
        Save a universal format chunk to the Format database (not the disk database)
        call save method to write changed chunks back to the disk database
        :param chunk: The chunk object to translate and save
        :return:
        """
        if not self.writeable:
            log.error("This object is not writeable")
            return
        try:
            self._commit_chunk(copy.deepcopy(chunk), *args)
        except Exception:
            log.error(f"Error saving chunk {chunk}", exc_info=True)
        self._changed = True

    def _commit_chunk(
        self, chunk: "Chunk", *args, recurse: bool = True,
    ):
        """
        Saves a universal amulet.api.chunk.Chunk object
        Calls the interface then the translator.
        It then calls _put_chunk_data to store the data returned by the interface
        """
        # get the coordinates for later
        cx, cz = chunk.cx, chunk.cz

        # Gets an interface, translator and most recent chunk version for the game version.
        interface, translator, chunk_version = self._get_interface_and_translator(
            self.max_world_version
        )

        chunk = self._convert_to_save(chunk, chunk_version, translator, recurse)
        chunk, chunk_palette = self._pack(chunk, translator, chunk_version)
        raw_chunk_data = self._encode(chunk, chunk_palette, interface)

        self._put_raw_chunk_data(cx, cz, raw_chunk_data, *args)

    def _convert_to_save(
        self,
        chunk: "Chunk",
        chunk_version: VersionNumberAny,
        translator: "Translator",
        recurse: bool = True,
    ) -> "Chunk":
        """Convert the Chunk in Universal format to a Chunk in the version specific format."""
        # create a new streamlined block block_palette and remap the data
        palette: List[numpy.ndarray] = []
        palette_len = 0
        for cy in chunk.blocks.sub_chunks:
            sub_chunk_palette, sub_chunk = numpy.unique(
                chunk.blocks.get_sub_chunk(cy), return_inverse=True
            )
            chunk.blocks.add_sub_chunk(
                cy, sub_chunk.astype(numpy.uint32).reshape((16, 16, 16)) + palette_len
            )
            palette_len += len(sub_chunk_palette)
            palette.append(sub_chunk_palette)

        if palette:
            chunk_palette, lut = numpy.unique(
                numpy.concatenate(palette), return_inverse=True
            )
            for cy in chunk.blocks.sub_chunks:
                chunk.blocks.add_sub_chunk(
                    cy, lut.astype(numpy.uint32)[chunk.blocks.get_sub_chunk(cy)]
                )
            chunk._block_palette = BlockManager(
                numpy.vectorize(chunk.block_palette.__getitem__)(chunk_palette)
            )
        else:
            chunk._block_palette = BlockManager()

        def get_chunk_callback(_: int, __: int) -> "Chunk":
            # conversion from universal should not require any data outside the block
            return chunk

        # translate from universal format to version format
        return translator.from_universal(
            chunk_version, self.translation_manager, chunk, get_chunk_callback, recurse,
        )

    def _pack(
        self, chunk: "Chunk", translator: "Translator", chunk_version: VersionNumberAny,
    ) -> Tuple["Chunk", AnyNDArray]:
        """Pack the chunk data into the format required by the encoder.
        This includes converting the string names to numerical formats for the versions that require it."""
        return translator.pack(chunk_version, self.translation_manager, chunk)

    def _encode(
        self, chunk: "Chunk", chunk_palette: AnyNDArray, interface: "Interface"
    ) -> Any:
        """Encode the data to the raw format as saved on disk."""
        return interface.encode(chunk, chunk_palette, self.max_world_version)

    def delete_chunk(self, cx: int, cz: int, *args):
        raise NotImplementedError

    def _put_raw_chunk_data(self, cx: int, cz: int, data: Any, *args):
        """
        Actually stores the data from the interface to disk.
        """
        raise NotImplementedError()

    def _get_raw_chunk_data(self, cx: int, cz: int, *args) -> Any:
        """
        Return the interface key and data to interface with given chunk coordinates.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :return: The raw chunk data.
        """
        raise NotImplementedError()



from __future__ import annotations

from typing import List, Tuple, Union, Iterable, Dict, TYPE_CHECKING
import numpy

import amulet_nbt as amulet_nbt

import amulet
from amulet.api.chunk import Chunk
from amulet.api.block import Block
from amulet.api.wrapper import Interface
from amulet.world_interface.chunk import translators
from amulet.api.data_types import AnyNDArray, SubChunkNDArray

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator
    from amulet.api.block_entity import BlockEntity
    from amulet.api.entity import Entity
    from amulet.api.chunk.blocks import Blocks


class BaseAnvilInterface(Interface):
    def __init__(self):
        feature_options = {
            "data_version": ["int"],  # int
            "last_update": ["long"],  # int
            "status": ["j13", "j14"],
            "light_populated": ["byte"],  # int
            "terrain_populated": ["byte"],  # int
            "V": ["byte"],  # int
            "inhabited_time": ["long"],  # int
            "biomes": ["256BA", "256IA", "1024IA"],  # Biomes
            "height_map": ["256IA", "C|36LA|V1", "C|36LA|V2", "C|36LA|V3", "C|36LA|V4"],
            # 'carving_masks': ['C|?BA'],
            "blocks": ["Sections|(Blocks,Data,Add)", "Sections|(BlockStates,Palette)"],
            "long_array_format": [
                "compact",
                "1.16",
            ],  # before the long array was just a bit stream but it is now separete longs. The upper bits are unused in some cases.
            "block_light": ["Sections|2048BA"],
            "sky_light": ["Sections|2048BA"],
            "block_entities": ["list"],
            "block_entity_format": ["namespace-str-id"],
            "block_entity_coord_format": ["xyz-int"],
            "entities": ["list"],
            "entity_format": ["namespace-str-id"],
            "entity_coord_format": ["Pos-list-double"],
            "tile_ticks": ["list", "list(optional)"],
            "liquid_ticks": ["list"],
            # 'lights': [],
            "liquids_to_be_ticked": ["16list|list"],
            "to_be_ticked": ["16list|list"],
            "post_processing": ["16list|list"],
            "structures": ["compound"],
        }
        self.features = {key: None for key in feature_options.keys()}

    def is_valid(self, key: Tuple) -> bool:
        return key[0] == "java" and self.minor_is_valid(key[1])

    @staticmethod
    def minor_is_valid(key: int):
        raise NotImplementedError

    def get_translator(
        self,
        max_world_version: Tuple[str, Union[int, Tuple[int, int, int]]],
        data: amulet_nbt.NBTFile = None,
    ) -> Tuple["Translator", int]:
        if data:
            data_version = data.get("DataVersion", amulet_nbt.TAG_Int(-1)).value
            key, version = (("java", data_version), data_version)
        else:
            key = max_world_version
            version = max_world_version[1]
        return translators.loader.get(key), version

    def decode(
        self, cx: int, cz: int, data: amulet_nbt.NBTFile
    ) -> Tuple["Chunk", AnyNDArray]:
        """
        Create an amulet.api.chunk.Chunk object from raw data given by the format.
        :param cx: chunk x coordinate
        :param cz: chunk z coordinate
        :param data: amulet_nbt.NBTFile
        :return: Chunk object in version-specific format, along with the block_palette for that chunk.
        """
        misc = {}
        chunk = Chunk(cx, cz)

        if self.features["last_update"] == "long":
            misc["last_update"] = (
                data["Level"].get("LastUpdate", amulet_nbt.TAG_Long(0)).value
            )

        if self.features["status"] in ["j13", "j14"]:
            chunk.status = data["Level"]["Status"].value
        else:
            status = "empty"
            if (
                self.features["terrain_populated"] == "byte"
                and data["Level"].get("TerrainPopulated", amulet_nbt.TAG_Byte()).value
            ):
                status = "decorated"
            if (
                self.features["light_populated"] == "byte"
                and data["Level"].get("LightPopulated", amulet_nbt.TAG_Byte()).value
            ):
                status = "postprocessed"

            chunk.status = status

        if self.features["V"] == "byte":
            misc["V"] = data["Level"]["V"].value

        if self.features["inhabited_time"] == "long":
            misc["inhabited_time"] = (
                data["Level"].get("InhabitedTime", amulet_nbt.TAG_Long(0)).value
            )

        if self.features["biomes"] is not None:
            if "Biomes" in data["Level"]:
                biomes = data["Level"]["Biomes"].value
                if self.features["biomes"] in ["256BA", "256IA"]:
                    chunk.biomes = biomes.astype(numpy.uint32).reshape((16, 16))
                elif self.features["biomes"] == "1024IA":
                    chunk.biomes = {
                        sy: arr
                        for sy, arr in enumerate(
                            numpy.split(
                                numpy.transpose(
                                    biomes.astype(numpy.uint32).reshape(64, 4, 4),
                                    (2, 0, 1),
                                ),  # YZX -> XYZ
                                16,
                                1,
                            )
                        )
                    }

        if self.features["height_map"] == "256IA":
            misc["height_map256IA"] = data["Level"]["HeightMap"].value
        elif self.features["height_map"] in [
            "C|36LA|V1",
            "C|36LA|V2",
            "C|36LA|V3",
            "C|36LA|V4",
        ]:
            if "Heightmaps" in data["Level"]:
                misc["height_mapC|36LA"] = data["Level"]["Heightmaps"]

        if "Sections" in data["Level"]:
            if self.features["blocks"] in [
                "Sections|(Blocks,Data,Add)",
                "Sections|(BlockStates,Palette)",
            ]:
                chunk.blocks, palette = self._decode_blocks(data["Level"]["Sections"])
            else:
                raise Exception(f'Unsupported block format {self.features["blocks"]}')

            if self.features["block_light"] == "Sections|2048BA":
                misc["block_light"] = {
                    section["Y"].value: section["BlockLight"]
                    for section in data["Level"]["Sections"]
                    if "BlockLight" in section
                }

            if self.features["sky_light"] == "Sections|2048BA":
                misc["sky_light"] = {
                    section["Y"].value: section["SkyLight"]
                    for section in data["Level"]["Sections"]
                    if "SkyLight" in section
                }
        else:
            palette = numpy.array([Block(namespace="minecraft", base_name="air")])

        if self.features["entities"] == "list":
            if amulet.entity_support:
                chunk.entities = self._decode_entities(
                    data["Level"].get("Entities", amulet_nbt.TAG_List())
                )
            else:
                misc["java_entities_temp"] = self._decode_entities(
                    data["Level"].get("Entities", amulet_nbt.TAG_List())
                )

        if self.features["block_entities"] == "list":
            chunk.block_entities = self._decode_block_entities(
                data["Level"].get("TileEntities", amulet_nbt.TAG_List())
            )

        if self.features["tile_ticks"] == "list":
            misc["tile_ticks"] = data["Level"].get("TileTicks", amulet_nbt.TAG_List())

        if self.features["liquid_ticks"] == "list":
            if "LiquidTicks" in data["Level"]:
                misc["liquid_ticks"] = data["Level"]["LiquidTicks"]

        if self.features["liquids_to_be_ticked"] == "16list|list":
            if "LiquidsToBeTicked" in data["Level"]:
                misc["liquids_to_be_ticked"] = data["Level"]["LiquidsToBeTicked"]

        if self.features["to_be_ticked"] == "16list|list":
            if "ToBeTicked" in data["Level"]:
                misc["to_be_ticked"] = data["Level"]["ToBeTicked"]

        if self.features["post_processing"] == "16list|list":
            if "PostProcessing" in data["Level"]:
                misc["post_processing"] = data["Level"]["PostProcessing"]

        if self.features["structures"] == "compound":
            if "Structures" in data["Level"]:
                misc["structures"] = data["Level"]["Structures"]

        chunk.misc = misc

        return chunk, palette

    def encode(
        self, chunk: "Chunk", palette: AnyNDArray, max_world_version: Tuple[str, int]
    ) -> amulet_nbt.NBTFile:
        """
        Encode a version-specific chunk to raw data for the format to store.
        :param chunk: The version-specific chunk to translate and encode.
        :param palette: The block_palette the ids in the chunk correspond to.
        :param max_world_version: The key to use to find the encoder.
        :return: amulet_nbt.NBTFile
        """

        misc = chunk.misc
        data = amulet_nbt.NBTFile(amulet_nbt.TAG_Compound(), "")
        data["Level"] = amulet_nbt.TAG_Compound()
        data["Level"]["xPos"] = amulet_nbt.TAG_Int(chunk.cx)
        data["Level"]["zPos"] = amulet_nbt.TAG_Int(chunk.cz)
        if self.features["data_version"] == "int":
            data["DataVersion"] = amulet_nbt.TAG_Int(max_world_version[1])

        if self.features["last_update"] == "long":
            data["Level"]["LastUpdate"] = amulet_nbt.TAG_Long(
                misc.get("last_update", 0)
            )

        # Order the float value based on the order they would be run. Newer replacements for the same come just after
        # to save back find the next lowest valid value.
        if self.features["status"] in ["j13", "j14"]:
            status = chunk.status.as_type(self.features["status"])
            data["Level"]["Status"] = amulet_nbt.TAG_String(status)

        else:
            status = chunk.status.as_type("float")
            if self.features["terrain_populated"] == "byte":
                data["Level"]["TerrainPopulated"] = amulet_nbt.TAG_Byte(
                    int(status > -0.3)
                )

            if self.features["light_populated"] == "byte":
                data["Level"]["LightPopulated"] = amulet_nbt.TAG_Byte(
                    int(status > -0.2)
                )

        if self.features["V"] == "byte":
            data["Level"]["V"] = amulet_nbt.TAG_Byte(misc.get("V", 1))

        if self.features["inhabited_time"] == "long":
            data["Level"]["InhabitedTime"] = amulet_nbt.TAG_Long(
                misc.get("inhabited_time", 0)
            )

        if self.features["biomes"] == "256BA":  # TODO: support the optional variant
            if chunk.status.value > -0.7:
                chunk.biomes.convert_to_2d()
                data["Level"]["Biomes"] = amulet_nbt.TAG_Byte_Array(
                    chunk.biomes.astype(dtype=numpy.uint8)
                )
        elif self.features["biomes"] == "256IA":
            if chunk.status.value > -0.7:
                chunk.biomes.convert_to_2d()
                data["Level"]["Biomes"] = amulet_nbt.TAG_Int_Array(
                    chunk.biomes.astype(dtype=numpy.uint32)
                )
        elif self.features["biomes"] == "1024IA":
            if chunk.status.value > -0.7:
                chunk.biomes.convert_to_3d()
                data["Level"]["Biomes"] = amulet_nbt.TAG_Int_Array(
                    numpy.transpose(
                        numpy.asarray(chunk.biomes[:, 0:64, :]).astype(numpy.uint32),
                        (1, 2, 0),
                    ).ravel(),  # YZX -> XYZ
                )

        if self.features["height_map"] == "256IA":
            data["Level"]["HeightMap"] = amulet_nbt.TAG_Int_Array(
                misc.get("height_map256IA", numpy.zeros(256, dtype=numpy.uint32))
            )
        elif self.features["height_map"] in [
            "C|36LA|V1",
            "C|36LA|V2",
            "C|36LA|V3",
            "C|36LA|V4",
        ]:
            maps = [
                "WORLD_SURFACE_WG",
                "OCEAN_FLOOR_WG",
                "MOTION_BLOCKING",
                "MOTION_BLOCKING_NO_LEAVES",
                "OCEAN_FLOOR",
            ]
            if self.features["height_map"] == "C|36LA|V1":  # 1466
                maps = ("LIQUID", "SOILD", "LIGHT", "RAIN")
            elif self.features["height_map"] == "C|36LA|V2":  # 1484
                maps.append("LIGHT_BLOCKING")
            elif self.features["height_map"] == "C|36LA|V3":  # 1503
                maps.append("LIGHT_BLOCKING")
                maps.append("WORLD_SURFACE")
            elif self.features["height_map"] == "C|36LA|V4":  # 1908
                maps.append("WORLD_SURFACE")
            else:
                raise Exception
            heightmaps = misc.get("height_mapC|36LA", amulet_nbt.TAG_Compound())
            for heightmap in maps:
                if heightmap not in heightmaps:
                    heightmaps[heightmap] = amulet_nbt.TAG_Long_Array(
                        numpy.zeros(36, dtype=">i8")
                    )
            data["Level"]["Heightmaps"] = heightmaps

        if self.features["blocks"] in [
            "Sections|(Blocks,Data,Add)",
            "Sections|(BlockStates,Palette)",
        ]:
            data["Level"]["Sections"] = self._encode_blocks(chunk.blocks, palette)
        else:
            raise Exception(f'Unsupported block format {self.features["blocks"]}')

        if self.features["block_light"] in ["Sections|2048BA"] or self.features[
            "sky_light"
        ] in ["Sections|2048BA"]:
            for section in data["Level"]["Sections"]:
                y = section["Y"].value
                if self.features["block_light"] == "Sections|2048BA":
                    block_light = misc.get("block_light", {})
                    if y in block_light:
                        section["BlockLight"] = block_light[y]
                    else:
                        section["BlockLight"] = amulet_nbt.TAG_Byte_Array(
                            numpy.full(2048, 255, dtype=numpy.uint8)
                        )
                if self.features["sky_light"] == "Sections|2048BA":
                    sky_light = misc.get("sky_light", {})
                    if y in sky_light:
                        section["SkyLight"] = sky_light[y]
                    else:
                        section["SkyLight"] = amulet_nbt.TAG_Byte_Array(
                            numpy.full(2048, 255, dtype=numpy.uint8)
                        )

        if self.features["entities"] == "list":
            if amulet.entity_support:
                data["Level"]["Entities"] = self._encode_entities(chunk.entities)
            else:
                data["Level"]["Entities"] = self._encode_entities(
                    misc.get("java_entities_temp", amulet_nbt.TAG_List())
                )

        if self.features["block_entities"] == "list":
            data["Level"]["TileEntities"] = self._encode_block_entities(
                chunk.block_entities
            )

        if self.features["tile_ticks"] in ["list", "list(optional)"]:
            ticks = misc.get("tile_ticks", amulet_nbt.TAG_List())
            if self.features["tile_ticks"] == "list(optional)":
                if len(ticks) > 0:
                    data["Level"]["TileTicks"] = ticks
            elif self.features["tile_ticks"] == "list":
                data["Level"]["TileTicks"] = ticks

        if self.features["liquid_ticks"] == "list":
            data["Level"]["LiquidTicks"] = misc.get(
                "liquid_ticks", amulet_nbt.TAG_List()
            )

        if self.features["liquids_to_be_ticked"] == "16list|list":
            data["Level"]["LiquidsToBeTicked"] = misc.get(
                "liquids_to_be_ticked",
                amulet_nbt.TAG_List([amulet_nbt.TAG_List() for _ in range(16)]),
            )

        if self.features["to_be_ticked"] == "16list|list":
            data["Level"]["ToBeTicked"] = misc.get(
                "to_be_ticked",
                amulet_nbt.TAG_List([amulet_nbt.TAG_List() for _ in range(16)]),
            )

        if self.features["post_processing"] == "16list|list":
            data["Level"]["PostProcessing"] = misc.get(
                "post_processing",
                amulet_nbt.TAG_List([amulet_nbt.TAG_List() for _ in range(16)]),
            )

        if self.features["structures"] == "compound":
            data["Level"]["Structures"] = misc.get(
                "structures",
                amulet_nbt.TAG_Compound(
                    {
                        "References": amulet_nbt.TAG_Compound(),
                        "Starts": amulet_nbt.TAG_Compound(),
                    }
                ),
            )

        return data

    def _decode_entities(self, entities: amulet_nbt.TAG_List) -> List["Entity"]:
        entities_out = []
        for nbt in entities:
            entity = self._decode_entity(
                amulet_nbt.NBTFile(nbt),
                self.features["entity_format"],
                self.features["entity_coord_format"],
            )
            if entity is not None:
                entities_out.append(entity)

        return entities_out

    def _encode_entities(self, entities: Iterable["Entity"]) -> amulet_nbt.TAG_List:
        entities_out = []
        for entity in entities:
            nbt = self._encode_entity(
                entity,
                self.features["entity_format"],
                self.features["entity_coord_format"],
            )
            if nbt is not None:
                entities_out.append(nbt.value)

        return amulet_nbt.TAG_List(entities_out)

    def _decode_block_entities(
        self, block_entities: amulet_nbt.TAG_List
    ) -> List["BlockEntity"]:
        entities_out = []
        for nbt in block_entities:
            entity = self._decode_block_entity(
                amulet_nbt.NBTFile(nbt),
                self.features["block_entity_format"],
                self.features["block_entity_coord_format"],
            )
            if entity is not None:
                entities_out.append(entity)

        return entities_out

    def _encode_block_entities(
        self, block_entities: Iterable["BlockEntity"]
    ) -> amulet_nbt.TAG_List:
        entities_out = []
        for entity in block_entities:
            nbt = self._encode_block_entity(
                entity,
                self.features["block_entity_format"],
                self.features["block_entity_coord_format"],
            )
            if nbt is not None:
                entities_out.append(nbt.value)

        return amulet_nbt.TAG_List(entities_out)

    def _decode_blocks(
        self, chunk_sections: amulet_nbt.TAG_List
    ) -> Tuple[Dict[int, SubChunkNDArray], AnyNDArray]:
        raise NotImplementedError

    def _encode_blocks(
        self, blocks: "Blocks", palette: AnyNDArray
    ) -> amulet_nbt.TAG_List:
        raise NotImplementedError



import os
from typing import Optional, Union, List, Tuple, Dict, Generator, TYPE_CHECKING
import numpy

from amulet import log
from amulet.api.data_types import (
    AnyNDArray,
    VersionNumberAny,
    PathOrBuffer,
)
from amulet.api.registry import BlockManager
from amulet.api.wrapper import StructureFormatWrapper
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.errors import ObjectReadError, ObjectWriteError, ChunkDoesNotExist

from .construction import ConstructionWriter, ConstructionReader, ConstructionSection
from .interface import Construction0Interface, ConstructionInterface

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator, Interface

construction_0_interface = Construction0Interface()


class ConstructionFormatWrapper(StructureFormatWrapper):
    def __init__(self, path: PathOrBuffer, mode: str = "r"):
        super().__init__(path)
        assert mode in ("r", "w"), 'Mode must be either "r" or "w".'
        self._mode = mode
        if isinstance(path, str):
            assert path.endswith(".construction"), 'Path must end with ".construction"'
            if mode == "r":
                assert os.path.isfile(path), "File specified does not exist."
        self._data: Optional[Union[ConstructionWriter, ConstructionReader]] = None
        self._open = False
        self._platform = "java"
        self._version = (1, 15, 2)
        self._selection: SelectionGroup = SelectionGroup()

        # used to look up which sections are in a given chunk when loading
        self._chunk_to_section: Optional[Dict[Tuple[int, int], List[int]]] = None

        # used to look up which selection boxes intersect a given chunk (boxes are clipped to the size of the chunk)
        self._chunk_to_box: Optional[Dict[Tuple[int, int], List[SelectionBox]]] = None

    @property
    def readable(self) -> bool:
        """Can this object have data read from it."""
        return self._mode == "r"

    @property
    def writeable(self) -> bool:
        """Can this object have data written to it."""
        return self._mode == "w"

    @staticmethod
    def is_valid(path: PathOrBuffer) -> bool:
        """
        Returns whether this format is able to load the given object.

        :param path: The path of the object to load.
        :return: True if the world can be loaded by this format, False otherwise.
        """
        return os.path.isfile(path) and path.endswith(".construction")

    @property
    def platform(self) -> str:
        """Platform string ("bedrock" / "java" / ...)"""
        return self._platform

    @platform.setter
    def platform(self, platform: str):
        if self._open:
            log.error(
                "Construction platform cannot be changed after the object has been opened."
            )
            return
        self._platform = platform

    @property
    def version(self) -> Tuple[int, int, int]:
        return self._version

    @version.setter
    def version(self, version: Tuple[int, int, int]):
        if self._open:
            log.error(
                "Construction version cannot be changed after the object has been opened."
            )
            return
        self._version = version

    @property
    def selection(self) -> SelectionGroup:
        """Platform string ("bedrock" / "java" / ...)"""
        return self._selection

    @selection.setter
    def selection(self, selection: SelectionGroup):
        if self._open:
            log.error(
                "Construction selection cannot be changed after the object has been opened."
            )
            return
        self._selection = selection

    def _get_interface(
        self, max_world_version, raw_chunk_data=None
    ) -> "ConstructionInterface":
        return construction_0_interface

    def _get_interface_and_translator(
        self, max_world_version, raw_chunk_data=None
    ) -> Tuple["Interface", "Translator", VersionNumberAny]:
        interface = self._get_interface(max_world_version, raw_chunk_data)
        translator, version_identifier = interface.get_translator(
            max_world_version, raw_chunk_data, self.translation_manager
        )
        return interface, translator, version_identifier

    def open(self):
        """Open the database for reading and writing"""
        if self._open:
            return
        if self._mode == "r":
            assert (
                isinstance(self.path_or_buffer, str)
                and os.path.isfile(self.path_or_buffer)
            ) or hasattr(self.path_or_buffer, "read"), "File specified does not exist."
            self._data = ConstructionReader(self.path_or_buffer)
            self._platform = self._data.source_edition
            self._version = self._data.source_version
            self._selection = SelectionGroup(
                [
                    SelectionBox((minx, miny, minz), (maxx, maxy, maxz))
                    for minx, miny, minz, maxx, maxy, maxz in self._data.selection
                ]
            )

            self._chunk_to_section = {}
            for index, (x, _, z, _, _, _, _, _) in enumerate(self._data.sections):
                cx = x >> 4
                cz = z >> 4
                self._chunk_to_section.setdefault((cx, cz), []).append(index)
        else:
            self._data = ConstructionWriter(
                self.path_or_buffer,
                self.platform,
                self.version,
                [box.bounds for box in self.selection.selection_boxes],
            )
            self._chunk_to_box = {}
            for box in self.selection.selection_boxes:
                for cx, cz in box.chunk_locations():
                    self._chunk_to_box.setdefault((cx, cz), []).append(
                        box.intersection(SelectionBox.create_chunk_box(cx, cz))
                    )
        self._open = True

    @property
    def has_lock(self) -> bool:
        """Verify that the world database can be read and written"""
        return True

    def save(self):
        """Save the data back to the disk database"""
        pass

    def close(self):
        """Close the disk database"""
        self._data.close()

    def unload(self):
        """Unload data stored in the Format class"""
        pass

    def all_chunk_coords(self, *args) -> Generator[Tuple[int, int], None, None]:
        """A generator of all chunk coords"""
        if self._mode == "r":
            yield from self._chunk_to_section.keys()
        else:
            raise ObjectReadError("all_chunk_coords is only valid in read mode")

    def _pack(
        self, chunk: "Chunk", translator: "Translator", chunk_version: VersionNumberAny,
    ) -> Tuple["Chunk", AnyNDArray]:
        return chunk, numpy.array(chunk.block_palette.blocks())

    def _encode(
        self, chunk: Chunk, chunk_palette: AnyNDArray, interface: ConstructionInterface,
    ):
        return interface.encode(
            chunk,
            chunk_palette,
            self.max_world_version,
            self._chunk_to_box.get((chunk.cx, chunk.cz)),
        )

    def _unpack(
        self,
        translator: "Translator",
        game_version: VersionNumberAny,
        chunk: "Chunk",
        chunk_palette: AnyNDArray,
    ) -> "Chunk":
        palette = chunk._block_palette = BlockManager()
        lut = numpy.array([palette.get_add_block(block) for block in chunk_palette])
        if len(palette.blocks()) != len(chunk_palette):
            # if a blockstate was defined twice
            for cy in chunk.blocks.sub_chunks:
                chunk.blocks.add_sub_chunk(cy, lut[chunk.blocks.get_sub_chunk(cy)])
        return chunk

    def delete_chunk(self, cx: int, cz: int, *args):
        raise ObjectWriteError(
            "delete_chunk is not a valid method for a construction file"
        )

    def _put_raw_chunk_data(
        self, cx: int, cz: int, data: List[ConstructionSection], *args
    ):
        """
        Actually stores the data from the interface to disk.
        """
        if self._mode == "w":
            for section in data:
                self._data.write(section)
        else:
            raise ObjectWriteError("The construction file is not open for writing.")

    def _get_raw_chunk_data(self, cx: int, cz: int, *args) -> List[ConstructionSection]:
        """
        Return the interface key and data to interface with given chunk coordinates.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :return: The raw chunk data.
        """
        if self._mode == "r":
            if (cx, cz) in self._chunk_to_section:
                return [
                    self._data.read(index) for index in self._chunk_to_section[(cx, cz)]
                ]
            else:
                raise ChunkDoesNotExist
        else:
            raise ObjectReadError("The construction file is not open for reading.")



from __future__ import annotations

import os
import struct
from typing import Type, Union, Tuple, IO, List, Optional

from amulet import Block
from amulet.api.registry import BlockManager
from amulet.api.entity import Entity
from amulet.api.block_entity import BlockEntity

import amulet_nbt

import numpy

INT_TRIPLET = Tuple[int, int, int]

INT_STRUCT = struct.Struct(">I")
SECTION_ENTRY_TYPE = numpy.dtype(
    [
        ("sx", "i4"),
        ("sy", "i4"),
        ("sz", "i4"),
        ("shapex", "i1"),
        ("shapey", "i1"),
        ("shapez", "i1"),
        ("position", "i4"),
        ("length", "i4"),
    ]
)

magic_num = b"constrct"
magic_num_len = len(magic_num)

max_format_version = 0
max_section_version = 0


class ConstructionSection:
    __slots__ = (
        "sx",
        "sy",
        "sz",
        "blocks",
        "palette",
        "shape",
        "entities",
        "block_entities",
    )

    def __init__(
        self,
        min_position: INT_TRIPLET,
        shape: INT_TRIPLET,
        blocks: Optional[numpy.ndarray],
        palette: List[Block],
        entities: List[Entity],
        block_entities: List[BlockEntity],
    ):
        self.sx, self.sy, self.sz = min_position
        self.shape = shape
        self.blocks = blocks
        if blocks is not None:
            assert isinstance(self.blocks, numpy.ndarray)
            assert (
                self.shape == self.blocks.shape
            ), "blocks shape does not match the specified section shape"
        self.palette = palette
        self.entities = entities
        self.block_entities = block_entities

    def __eq__(self, other):
        return (
            isinstance(other, ConstructionSection)
            and self.sx == other.sx
            and self.sy == other.sy
            and self.sz == other.sz
            and self.shape == other.shape
            and numpy.equal(self.blocks, other.blocks).all()
            and self.entities == other.entities
            and self.block_entities == other.block_entities
        )

    @property
    def location(self) -> Tuple[int, int, int]:
        return self.sx, self.sy, self.sz


class ConstructionReader:
    def __init__(self, file_or_buffer: Union[str, IO]):
        self._format_version: Optional[int] = None
        self._section_version: Optional[int] = None

        if isinstance(file_or_buffer, str):
            assert os.path.isfile(
                file_or_buffer
            ), f"There is no file located at {file_or_buffer}"
            self._buffer = open(file_or_buffer, "rb")
        else:
            assert hasattr(
                file_or_buffer, "read"
            ), "Construction file buffer does not have a read mode"
            self._buffer = file_or_buffer

        self._source_edition: Optional[str] = None
        self._source_version: Optional[INT_TRIPLET] = None

        self._selection_boxes: Optional[
            List[Tuple[int, int, int, int, int, int]]
        ] = None

        self._metadata: Optional[amulet_nbt.NBTFile] = None
        self._section_index_table: Optional[
            List[Tuple[int, int, int, int, int, int, int, int]]
        ] = None
        self._palette: Optional[List[Block]] = None
        self._init_read()

    @property
    def metadata(self) -> amulet_nbt.NBTFile:
        return self._metadata

    @property
    def palette(self) -> List[Block]:
        return self._palette.copy()

    @property
    def sections(self) -> List[Tuple[int, int, int, int, int, int, int, int]]:
        return self._section_index_table.copy()

    @property
    def source_edition(self) -> str:
        return self._source_edition

    @property
    def source_version(self) -> INT_TRIPLET:
        return self._source_version

    @property
    def selection(self) -> List[Tuple[int, int, int, int, int, int]]:
        return self._selection_boxes.copy()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @staticmethod
    def _unpack_palette(raw_palette: amulet_nbt.TAG_List) -> List[Block]:
        block_palette = []
        extra_block_map = {}
        for block_index, block_nbt in enumerate(raw_palette):
            block_nbt: amulet_nbt.TAG_Compound
            block_namespace = block_nbt["namespace"].value
            block_basename = block_nbt["blockname"].value
            block = Block(
                namespace=block_namespace,
                base_name=block_basename,
                properties=block_nbt["properties"].value,
            )

            if block_nbt["extra_blocks"].value:
                extra_block_map[block_index] = block_nbt["extra_blocks"].value

            block_palette.append(block)

        for block_index, extra_blocks in extra_block_map.items():
            extra_block_objects = [block_palette[i.value] for i in extra_blocks]

            resulting_block = block_palette[block_index]
            for extra_block in extra_block_objects:
                resulting_block = resulting_block + extra_block

            block_palette[block_index] = resulting_block
        return block_palette

    def _init_read(self):
        """data to be read at init in read mode"""
        magic_num_1 = self._buffer.read(8)
        assert magic_num_1 == magic_num, f"This file is not a construction file."
        self._format_version = struct.unpack(">B", self._buffer.read(1))[0]
        if self._format_version == 0:
            self._buffer.seek(-magic_num_len, os.SEEK_END)
            magic_num_2 = self._buffer.read(8)
            assert (
                magic_num_2 == magic_num
            ), "It looks like this file is corrupt. It probably wasn't saved properly"

            self._buffer.seek(-magic_num_len - INT_STRUCT.size, os.SEEK_END)
            metadata_end = self._buffer.tell()
            metadata_start = INT_STRUCT.unpack(self._buffer.read(INT_STRUCT.size))[0]
            self._buffer.seek(metadata_start)

            metadata = amulet_nbt.load(
                buffer=self._buffer.read(metadata_end - metadata_start),
                compressed=True,
            )

            try:
                self._source_edition = metadata["export_version"]["edition"].value
                self._source_version = tuple(
                    map(lambda v: v.value, metadata["export_version"]["version"])
                )
            except KeyError as e:
                raise AssertionError(
                    f'Missing export version identifying key "{e.args[0]}"'
                )

            self._section_version = metadata["section_version"].value

            self._palette = self._unpack_palette(metadata["block_palette"])

            self._selection_boxes = (
                metadata["selection_boxes"].value.reshape(-1, 6).tolist()
            )

            self._section_index_table = (
                metadata["section_index_table"].value.view(SECTION_ENTRY_TYPE).tolist()
            )

        else:
            raise Exception(
                f"This wrapper doesn't support any construction version higher than {max_format_version}"
            )

    @staticmethod
    def _parse_entities(entities: amulet_nbt.TAG_List) -> List[Entity]:
        return [
            Entity(
                entity["namespace"].value,
                entity["base_name"].value,
                entity["x"].value,
                entity["y"].value,
                entity["z"].value,
                amulet_nbt.NBTFile(entity["nbt"]),
            )
            for entity in entities
        ]

    @staticmethod
    def _parse_block_entities(block_entities: amulet_nbt.TAG_List) -> List[BlockEntity]:
        return [
            BlockEntity(
                block_entity["namespace"].value,
                block_entity["base_name"].value,
                block_entity["x"].value,
                block_entity["y"].value,
                block_entity["z"].value,
                amulet_nbt.NBTFile(block_entity["nbt"]),
            )
            for block_entity in block_entities
        ]

    def read(self, section_index: int):
        if self._format_version == 0:
            (
                sx,
                sy,
                sz,
                shapex,
                shapey,
                shapez,
                position,
                length,
            ) = self._section_index_table[section_index]
            self._buffer.seek(position)
            nbt_obj = amulet_nbt.load(buffer=self._buffer.read(length))
            if nbt_obj["blocks_array_type"].value == -1:
                blocks = None
                block_entities = None
            else:
                blocks = numpy.reshape(
                    nbt_obj["blocks"].value, (shapex, shapey, shapez)
                )
                block_entities = self._parse_block_entities(nbt_obj["block_entities"])

            return ConstructionSection(
                (sx, sy, sz),
                (shapex, shapey, shapez),
                blocks,
                self._palette,
                self._parse_entities(nbt_obj["entities"]),
                block_entities,
            )
        else:
            raise Exception(
                f"This wrapper doesn't support any construction version higher than {max_format_version}"
            )

    def close(self):
        self._buffer.close()


class ConstructionWriter:
    def __init__(
        self,
        file_or_buffer: Union[str, IO],
        source_edition: str,
        source_version: INT_TRIPLET,
        selection_boxes: Optional[List[Tuple[int, int, int, int, int, int]]] = None,
        format_version: int = max_format_version,
        section_version: int = max_section_version,
    ):
        assert (
            format_version <= max_format_version
        ), f"This construction writer does not support format versions above {max_format_version}"
        assert (
            section_version <= max_section_version
        ), f"This construction writer does not support section versions above {max_section_version}"
        self._format_version = format_version
        self._section_version = section_version

        if isinstance(file_or_buffer, str):
            self._buffer = open(file_or_buffer, "wb")
        else:
            assert hasattr(
                file_or_buffer, "write"
            ), "Construction file buffer does not have a write mode"
            self._buffer = file_or_buffer

        self._source_edition = source_edition
        self._source_version = source_version

        self._selection_boxes = selection_boxes or []

        self._metadata: Optional[amulet_nbt.NBTFile] = None
        self._section_index_table: List[
            Tuple[int, int, int, int, int, int, int, int]
        ] = []
        self._palette: BlockManager = BlockManager()
        self._init_write()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _init_write(self):
        """data to be written at init in write mode"""
        self._buffer.write(magic_num)
        self._buffer.write(struct.pack(">B", self._format_version))
        if self._format_version == 0:
            self._metadata = amulet_nbt.NBTFile(
                amulet_nbt.TAG_Compound(
                    {
                        "created_with": amulet_nbt.TAG_String("amulet_python_wrapper"),
                        "selection_boxes": amulet_nbt.TAG_Int_Array(
                            [
                                c
                                for box in self._selection_boxes
                                if len(box) == 6
                                and all(isinstance(c, int) for c in box)
                                for c in box
                            ]
                        ),
                        "section_version": amulet_nbt.TAG_Byte(self._section_version),
                        "export_version": amulet_nbt.TAG_Compound(
                            {
                                "edition": amulet_nbt.TAG_String(self._source_edition),
                                "version": amulet_nbt.TAG_List(
                                    [
                                        amulet_nbt.TAG_Int(v)
                                        for v in self._source_version
                                    ]
                                ),
                            }
                        ),
                    }
                )
            )
        else:
            raise Exception(
                f"This wrapper doesn't support any construction version higher than {max_format_version}"
            )

    @staticmethod
    def _generate_block_entry(
        _block: Block, _palette_len, _extra_blocks
    ) -> amulet_nbt.TAG_Compound:
        return amulet_nbt.TAG_Compound(
            {
                "namespace": amulet_nbt.TAG_String(_block.namespace),
                "blockname": amulet_nbt.TAG_String(_block.base_name),
                "properties": amulet_nbt.TAG_Compound(_block.properties),
                "extra_blocks": amulet_nbt.TAG_List(
                    [
                        amulet_nbt.TAG_Int(
                            _palette_len + _extra_blocks.index(_extra_block)
                        )
                        for _extra_block in _block.extra_blocks
                    ]
                ),
            }
        )

    @staticmethod
    def _serialise_entities(entities: List[Entity]) -> amulet_nbt.TAG_List:
        return amulet_nbt.TAG_List(
            [
                amulet_nbt.TAG_Compound(
                    {
                        "namespace": amulet_nbt.TAG_String(entity.namespace),
                        "base_name": amulet_nbt.TAG_String(entity.base_name),
                        "x": amulet_nbt.TAG_Double(entity.x),
                        "y": amulet_nbt.TAG_Double(entity.y),
                        "z": amulet_nbt.TAG_Double(entity.z),
                        "nbt": entity.nbt.value,
                    }
                )
                for entity in entities
            ]
        )

    @staticmethod
    def _serialise_block_entities(
        block_entities: List[BlockEntity],
    ) -> amulet_nbt.TAG_List:
        return amulet_nbt.TAG_List(
            [
                amulet_nbt.TAG_Compound(
                    {
                        "namespace": amulet_nbt.TAG_String(block_entity.namespace),
                        "base_name": amulet_nbt.TAG_String(block_entity.base_name),
                        "x": amulet_nbt.TAG_Int(block_entity.x),
                        "y": amulet_nbt.TAG_Int(block_entity.y),
                        "z": amulet_nbt.TAG_Int(block_entity.z),
                        "nbt": block_entity.nbt.value,
                    }
                )
                for block_entity in block_entities
            ]
        )

    def _pack_palette(self) -> amulet_nbt.TAG_List:
        block_palette_nbt = amulet_nbt.TAG_List()
        extra_blocks = set()

        for block in self._palette.blocks():
            if len(block.extra_blocks) > 0:
                extra_blocks.update(block.extra_blocks)

        extra_blocks = list(extra_blocks)

        palette_len = len(self._palette)
        for block_entry in self._palette.blocks():
            block_palette_nbt.append(
                self._generate_block_entry(block_entry, palette_len, extra_blocks)
            )

        for extra_block in extra_blocks:
            block_palette_nbt.append(
                self._generate_block_entry(extra_block, palette_len, extra_blocks)
            )
        return block_palette_nbt

    def _exit_write(self):
        """data to be written at close in write mode"""
        if self._format_version == 0:
            metadata_start = self._buffer.tell()
            self._metadata["section_index_table"] = amulet_nbt.TAG_Byte_Array(
                numpy.array(self._section_index_table, dtype=SECTION_ENTRY_TYPE).view(
                    numpy.int8
                )
            )
            self._metadata["block_palette"] = self._pack_palette()
            self._metadata.save_to(self._buffer)
            self._buffer.write(INT_STRUCT.pack(metadata_start))
            self._buffer.write(magic_num)
        else:
            raise Exception(
                f"This wrapper doesn't support any construction version higher than {max_format_version}"
            )
        self._buffer.close()

    @staticmethod
    def _find_fitting_array_type(
        array: numpy.ndarray,
    ) -> Type[
        Union[
            amulet_nbt.TAG_Int_Array,
            amulet_nbt.TAG_Byte_Array,
            amulet_nbt.TAG_Long_Array,
        ]
    ]:
        max_element = array.max(initial=0)

        if max_element <= 127:
            return amulet_nbt.TAG_Byte_Array
        elif max_element <= 2_147_483_647:
            return amulet_nbt.TAG_Int_Array
        else:
            return amulet_nbt.TAG_Long_Array

    def write(self, section: ConstructionSection):
        if self._section_version == 0:
            sx, sy, sz = section.location
            shapex, shapey, shapez = section.shape
            blocks = section.blocks
            entities = section.entities
            block_entities = section.block_entities
            palette = section.palette
            for point, shape in zip((sx, sy, sz), (shapex, shapey, shapez)):
                assert shape >= 0, "Shape must be positive"
                assert point + shape <= (
                    ((point >> 4) + 1) << 4
                ), "Section does not fit in a sub-chunk"
            position = self._buffer.tell()

            _tag = amulet_nbt.TAG_Compound(
                {"entities": self._serialise_entities(entities)}
            )

            if blocks is None:
                _tag["blocks_array_type"] = amulet_nbt.TAG_Byte(-1)
            else:
                flattened_array = blocks.ravel()
                index, flattened_array = numpy.unique(
                    flattened_array, return_inverse=True
                )
                palette = numpy.array(palette, dtype=object)[index]
                lut = numpy.vectorize(self._palette.get_add_block)(palette)
                flattened_array = lut[flattened_array]
                array_type = self._find_fitting_array_type(flattened_array)
                _tag["blocks_array_type"] = amulet_nbt.TAG_Byte(array_type().tag_id)
                _tag["blocks"] = array_type(flattened_array)
                _tag["block_entities"] = self._serialise_block_entities(
                    block_entities or []
                )

            amulet_nbt.NBTFile(_tag).save_to(self._buffer)

            length = self._buffer.tell() - position
            self._section_index_table.append(
                (sx, sy, sz, shapex, shapey, shapez, position, length)
            )
        else:
            raise Exception(
                f"This wrapper doesn't support any section version higher than {max_section_version}"
            )

    def close(self):
        self._exit_write()
        self._buffer.close()



from typing import Tuple, Any, Union, Generator, List, Optional, TYPE_CHECKING
import traceback
import os

from amulet import log, IMG_DIRECTORY
from amulet.api.data_types import ChunkCoordinates
from .format_wrapper import BaseFormatWrapper

if TYPE_CHECKING:
    from amulet.api.chunk import Chunk
    from amulet.api.data_types import Dimension

missing_world_icon = os.path.abspath(
    os.path.join(IMG_DIRECTORY, "missing_world_icon.png")
)


class WorldFormatWrapper(BaseFormatWrapper):
    _missing_world_icon = missing_world_icon

    def __init__(self, world_path: str):
        super().__init__()
        self._path = world_path
        self._world_image_path = missing_world_icon
        self._changed: bool = False

    @property
    def sub_chunk_size(self) -> int:
        return 16

    @property
    def chunk_size(self) -> Tuple[int, Union[int, None], int]:
        return self.sub_chunk_size, self.sub_chunk_size * 16, self.sub_chunk_size

    @property
    def path(self) -> str:
        """The path to the world directory"""
        return self._path

    @property
    def world_name(self) -> str:
        """The name of the world"""
        return "Unknown World"

    @world_name.setter
    def world_name(self, value: str):
        raise NotImplementedError

    @property
    def last_played(self) -> int:
        raise NotImplementedError

    @property
    def game_version_string(self) -> str:
        raise NotImplementedError

    @property
    def world_path(self) -> str:
        """The path to the world directory"""
        log.info(
            f"Format.world_path is depreciated. Please used Format.path{traceback.format_exc()}"
        )
        return self._path

    @property
    def world_image_path(self) -> str:
        """The path to the world icon"""
        return self._world_image_path

    @property
    def dimensions(self) -> List["Dimension"]:
        """A list of all the dimensions contained in the world"""
        raise NotImplementedError

    def register_dimension(
        self, dimension_internal: Any, dimension_name: Optional["Dimension"] = None
    ):
        """
        Register a new dimension.
        :param dimension_internal: The internal representation of the dimension
        :param dimension_name: The name of the dimension shown to the user
        :return:
        """
        raise NotImplementedError

    def all_chunk_coords(
        self, dimension: "Dimension"
    ) -> Generator[ChunkCoordinates, None, None]:
        """A generator of all chunk coords in the given dimension"""
        raise NotImplementedError

    def load_chunk(self, cx: int, cz: int, dimension: "Dimension") -> "Chunk":
        """
        Loads and creates a universal amulet.api.chunk.Chunk object from chunk coordinates.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: dimension
        :return: The chunk at the given coordinates.
        """
        return super().load_chunk(cx, cz, dimension)

    def commit_chunk(self, chunk: "Chunk", dimension: "Dimension"):
        """
        Save a universal format chunk to the Format database (not the disk database)
        call save method to write changed chunks back to the disk database
        :param chunk: The chunk object to translate and save
        :param dimension: optional dimension
        :return:
        """
        super().commit_chunk(chunk, dimension)

    def delete_chunk(self, cx: int, cz: int, dimension: "Dimension"):
        raise NotImplementedError

    def _put_raw_chunk_data(
        self, cx: int, cz: int, data: Any, dimension: "Dimension", *args
    ):
        """
        Actually stores the data from the interface to disk.
        """
        raise NotImplementedError()

    def _get_raw_chunk_data(
        self, cx: int, cz: int, dimension: "Dimension", *args
    ) -> Any:
        """
        Return the interface key and data to interface with given chunk coordinates.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :return: The interface key for the get_interface method and the data to interface with.
        """
        raise NotImplementedError()



from __future__ import annotations

# from copy import deepcopy
import os
import shutil
from typing import Union, Generator, Dict, Optional, Tuple, Callable, Any, TYPE_CHECKING
from types import GeneratorType
import warnings

from amulet import log
from amulet.api.base_structure import BaseStructure
from amulet.api.block import Block
from amulet.api.block_entity import BlockEntity
from amulet.api.entity import Entity
from amulet.api.registry import BlockManager
from amulet.api.registry.biome_manager import BiomeManager
from amulet.api.errors import ChunkDoesNotExist, ChunkLoadError, LevelDoesNotExist
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.paths import get_temp_dir
from amulet.api.data_types import (
    OperationType,
    Dimension,
    DimensionCoordinates,
    VersionIdentifierType,
)
from amulet.utils.world_utils import block_coords_to_chunk_coords
from .chunk_manager import ChunkManager
from amulet.api.history.history_manager import MetaHistoryManager

if TYPE_CHECKING:
    from PyMCTranslate import TranslationManager
    from amulet.api.wrapper.world_format_wrapper import WorldFormatWrapper

ChunkCache = Dict[DimensionCoordinates, Optional[Chunk]]


class World(BaseStructure):
    """
    Class that handles world editing of any world format via an separate and flexible data format
    """

    def __init__(
        self, directory: str, world_wrapper: "WorldFormatWrapper", temp_dir: str = None
    ):
        self._directory = directory
        if temp_dir is None:
            self._temp_directory = get_temp_dir(self._directory)
        else:
            self._temp_directory = temp_dir

        self._world_wrapper = world_wrapper
        self._world_wrapper.open()

        self._block_palette = BlockManager()
        self._block_palette.get_add_block(
            Block(namespace="universal_minecraft", base_name="air")
        )  # ensure that index 0 is always air

        self._biome_palette = BiomeManager()
        self._biome_palette.get_add_biome("universal_minecraft:plains")

        self._history_manager = MetaHistoryManager()

        self._chunks: ChunkManager = ChunkManager(
            os.path.join(self._temp_directory, "chunks"), self
        )

        self._history_manager.register(self._chunks, True)

    @property
    def world_path(self) -> str:
        """The directory where the world is located"""
        return self._directory

    @property
    def changed(self) -> bool:
        """Has any data been modified but not saved to disk"""
        return self._history_manager.changed or self._world_wrapper.changed

    @property
    def history_manager(self) -> MetaHistoryManager:
        return self._history_manager

    @property
    def chunk_history_manager(self) -> ChunkManager:
        """A class storing previous versions of chunks to roll back to as required."""
        warnings.warn(
            "World.chunk_history_manager is depreciated and will be removed in the future. Please use World.chunks instead",
            DeprecationWarning,
        )
        return self._chunks

    def create_undo_point(self):
        """Create a restore point for all chunks that have changed."""
        self._history_manager.create_undo_point()

    @property
    def sub_chunk_size(self) -> int:
        """The normal dimensions of the chunk"""
        return self._world_wrapper.sub_chunk_size

    @property
    def chunk_size(self) -> Tuple[int, Union[int, None], int]:
        """The normal dimensions of the chunk"""
        return self._world_wrapper.chunk_size

    @property
    def translation_manager(self) -> "TranslationManager":
        """An instance of the translation class for use with this world."""
        return self._world_wrapper.translation_manager

    @property
    def world_wrapper(self) -> "WorldFormatWrapper":
        """A class to access data directly from the world."""
        return self._world_wrapper

    @property
    def palette(self) -> BlockManager:
        """The manager for the universal blocks in this world. New blocks must be registered here before adding to the world."""
        warnings.warn(
            "World.palette is depreciated and will be removed in the future. Please use World.block_palette instead",
            DeprecationWarning,
        )
        return self.block_palette

    @property
    def block_palette(self) -> BlockManager:
        """The manager for the universal blocks in this world. New blocks must be registered here before adding to the world."""
        return self._block_palette

    @property
    def biome_palette(self) -> BiomeManager:
        """The manager for the universal blocks in this world. New blocks must be registered here before adding to the world."""
        return self._biome_palette

    def save(
        self,
        wrapper: "WorldFormatWrapper" = None,
        progress_callback: Callable[[int, int], None] = None,
    ):
        """Save the world using the given wrapper.
        Leave as None to save back to the input wrapper.
        Optional progress callback to let the calling program know the progress. Input format chunk_index, chunk_count"""
        for chunk_index, chunk_count in self.save_iter(wrapper):
            if progress_callback is not None:
                progress_callback(chunk_index, chunk_count)

    def save_iter(
        self, wrapper: "WorldFormatWrapper" = None
    ) -> Generator[Tuple[int, int], None, None]:
        """Save the world using the given wrapper.
        Leave as None to save back to the input wrapper."""
        chunk_index = 0

        changed_chunks = list(self._chunks.changed_chunks())
        chunk_count = len(changed_chunks)

        if wrapper is None:
            wrapper = self._world_wrapper

        output_dimension_map = wrapper.dimensions

        # perhaps make this check if the directory is the same rather than if the class is the same
        save_as = wrapper is not self._world_wrapper
        if save_as:
            # The input wrapper is not the same as the loading wrapper (save-as)
            # iterate through every chunk in the input world and save them to the wrapper
            log.info(
                f"Converting world {self._world_wrapper.path} to world {wrapper.path}"
            )
            wrapper.translation_manager = (
                self._world_wrapper.translation_manager
            )  # TODO: this might cause issues in the future
            for dimension in self._world_wrapper.dimensions:
                chunk_count += len(
                    list(self._world_wrapper.all_chunk_coords(dimension))
                )

            for dimension in self._world_wrapper.dimensions:
                try:
                    if dimension not in output_dimension_map:
                        continue
                    for cx, cz in self._world_wrapper.all_chunk_coords(dimension):
                        log.info(f"Converting chunk {dimension} {cx}, {cz}")
                        try:
                            chunk = self._world_wrapper.load_chunk(cx, cz, dimension)
                            wrapper.commit_chunk(chunk, dimension)
                        except ChunkLoadError:
                            log.info(f"Error loading chunk {cx} {cz}", exc_info=True)
                        chunk_index += 1
                        yield chunk_index, chunk_count
                        if not chunk_index % 10000:
                            wrapper.save()
                            self._world_wrapper.unload()
                            wrapper.unload()
                except LevelDoesNotExist:
                    continue

        for dimension, cx, cz in changed_chunks:
            if dimension not in output_dimension_map:
                continue
            try:
                chunk = self.get_chunk(cx, cz, dimension)
            except ChunkDoesNotExist:
                wrapper.delete_chunk(cx, cz, dimension)
            except ChunkLoadError:
                pass
            else:
                wrapper.commit_chunk(chunk, dimension)
                chunk.changed = False
            chunk_index += 1
            yield chunk_index, chunk_count
            if not chunk_index % 10000:
                wrapper.save()
                wrapper.unload()

        self._history_manager.mark_saved()
        log.info(f"Saving changes to world {wrapper.path}")
        wrapper.save()
        log.info(f"Finished saving changes to world {wrapper.path}")

    def close(self):
        """Close the attached world and remove temporary files
        Use changed method to check if there are any changes that should be saved before closing."""
        # TODO: add "unsaved changes" check before exit
        shutil.rmtree(self._temp_directory, ignore_errors=True)
        self._world_wrapper.close()

    def unload(self, safe_area: Optional[Tuple[Dimension, int, int, int, int]] = None):
        """Unload all chunks not in the safe area
        Safe area format: dimension, min chunk X|Z, max chunk X|Z"""
        self._chunks.unload(safe_area)
        self._world_wrapper.unload()

    def get_chunk(self, cx: int, cz: int, dimension: Dimension) -> Chunk:
        """
        Gets the chunk data of the specified chunk coordinates.
        If the chunk does not exist ChunkDoesNotExist is raised.
        If some other error occurs then ChunkLoadError is raised (this error will also catch ChunkDoesNotExist)

        :param cx: The X coordinate of the desired chunk
        :param cz: The Z coordinate of the desired chunk
        :param dimension: The dimension to get the chunk from
        :return: A Chunk object containing the data for the chunk
        :raises: `amulet.api.errors.ChunkDoesNotExist` if the chunk does not exist or `amulet.api.errors.ChunkLoadError` if the chunk failed to load for some reason. (This also includes it not existing)
        """
        return self._chunks.get_chunk(dimension, cx, cz)

    def create_chunk(self, cx: int, cz: int, dimension: Dimension) -> Chunk:
        chunk = Chunk(cx, cz)
        self.put_chunk(chunk, dimension)
        return chunk

    def put_chunk(self, chunk: Chunk, dimension: Dimension):
        """Add a chunk to the universal world database"""
        self._chunks.put_chunk(chunk, dimension)

    def delete_chunk(self, cx: int, cz: int, dimension: Dimension):
        """Delete a chunk from the universal world database"""
        self._chunks.delete_chunk(dimension, cx, cz)

    def get_block(self, x: int, y: int, z: int, dimension: Dimension) -> Block:
        """
        Gets the universal Block object at the specified coordinates

        :param x: The X coordinate of the desired block
        :param y: The Y coordinate of the desired block
        :param z: The Z coordinate of the desired block
        :param dimension: The dimension of the desired block
        :return: The universal Block object representation of the block at that location
        :raise: Raises ChunkDoesNotExist or ChunkLoadError if the chunk was not loaded.
        """
        cx, cz = block_coords_to_chunk_coords(x, z, chunk_size=self.chunk_size[0])
        offset_x, offset_z = x - 16 * cx, z - 16 * cz

        return self.get_chunk(cx, cz, dimension).get_block(offset_x, y, offset_z)

    def get_version_block(
        self,
        x: int,
        y: int,
        z: int,
        dimension: Dimension,
        version: VersionIdentifierType,
    ) -> Tuple[Union[Block, Entity], Optional[BlockEntity]]:
        """
        Get a block at the specified location and convert it to the format of the version specified
        Note the odd return format. In most cases this will return (Block, None) or (Block, BlockEntity)
        but in select cases like item frames may return (Entity, None)

        :param x: The X coordinate of the desired block
        :param y: The Y coordinate of the desired block
        :param z: The Z coordinate of the desired block
        :param dimension: The dimension of the desired block
        :param version: The version to get the block converted to.
        :return: The block at the given location converted to the `version` format. Note the odd return format.
        :raise: Raises ChunkDoesNotExist or ChunkLoadError if the chunk was not loaded.
        """
        cx, cz = block_coords_to_chunk_coords(x, z, chunk_size=self.sub_chunk_size)
        chunk = self.get_chunk(cx, cz, dimension)
        offset_x, offset_z = x - 16 * cx, z - 16 * cz

        output, extra_output, _ = self.translation_manager.get_version(
            *version
        ).block.from_universal(
            chunk.get_block(offset_x, y, offset_z), chunk.block_entities.get((x, y, z))
        )
        return output, extra_output

    def set_version_block(
        self,
        x: int,
        y: int,
        z: int,
        dimension: Dimension,
        version: VersionIdentifierType,
        block: Block,
        block_entity: BlockEntity,
    ):
        """
        Convert the block and block_entity from the given version format to the universal format and set at the location

        :param x: The X coordinate of the desired block
        :param y: The Y coordinate of the desired block
        :param z: The Z coordinate of the desired block
        :param dimension: The dimension of the desired block
        :param version: The version to get the block converted to.
        :param block:
        :param block_entity:
        :return: The block at the given location converted to the `version` format. Note the odd return format.
        :raise: Raises ChunkLoadError if the chunk was not loaded correctly.
        """
        cx, cz = block_coords_to_chunk_coords(x, z, chunk_size=self.sub_chunk_size)
        try:
            chunk = self.get_chunk(cx, cz, dimension)
        except ChunkDoesNotExist:
            chunk = self.create_chunk(cx, cz, dimension)
        offset_x, offset_z = x - 16 * cx, z - 16 * cz

        (
            universal_block,
            universal_block_entity,
            _,
        ) = self.translation_manager.get_version(*version).block.to_universal(
            block, block_entity
        )
        chunk.set_block(offset_x, y, offset_z, block),
        chunk.block_entities[(x, y, z)] = block_entity

    def get_chunk_boxes(
        self,
        selection: Union[SelectionGroup, SelectionBox],
        dimension: Dimension,
        create_missing_chunks=False,
    ) -> Generator[Tuple[Chunk, SelectionBox], None, None]:
        """Given a selection will yield chunks and SubSelectionBoxes into that chunk

        :param selection: SelectionGroup or SelectionBox into the world
        :param dimension: The dimension to take effect in (defaults to overworld)
        :param create_missing_chunks: If a chunk does not exist an empty one will be created (defaults to false)
        """

        if isinstance(selection, SelectionBox):
            selection = SelectionGroup([selection])
        selection: SelectionGroup
        for (cx, cz), box in selection.sub_sections(self.sub_chunk_size):
            try:
                chunk = self.get_chunk(cx, cz, dimension)
            except ChunkDoesNotExist:
                if create_missing_chunks:
                    chunk = Chunk(cx, cz)
                    self.put_chunk(chunk, dimension)
                else:
                    continue
            except ChunkLoadError:
                continue

            yield chunk, box

    def get_chunk_slices(
        self,
        selection: Union[SelectionGroup, SelectionBox],
        dimension: Dimension,
        create_missing_chunks=False,
    ) -> Generator[Tuple[Chunk, Tuple[slice, slice, slice], SelectionBox], None, None]:
        """Given a selection will yield chunks, slices into that chunk and the corresponding box

        :param selection: SelectionGroup or SelectionBox into the world
        :param dimension: The dimension to take effect in (defaults to overworld)
        :param create_missing_chunks: If a chunk does not exist an empty one will be created (defaults to false)
        Usage:
        for chunk, slice, box in world.get_chunk_slices(selection):
            chunk.blocks[slice] = ...
        """
        for chunk, box in self.get_chunk_boxes(
            selection, dimension, create_missing_chunks
        ):
            slices = box.chunk_slice(chunk.cx, chunk.cz, self.sub_chunk_size)
            yield chunk, slices, box

    # def get_entities_in_box(
    #     self, box: "SelectionGroup"
    # ) -> Generator[Tuple[Coordinates, List[object]], None, None]:
    #     # TODO: some of this logic can probably be moved the chunk class and have this method call that
    #     # TODO: update this to use the newer entity API
    #     out_of_place_entities = []
    #     entity_map: Dict[Tuple[int, int], List[List[object]]] = {}
    #     for chunk, subbox in self.get_chunk_boxes(box):
    #         entities = chunk.entities
    #         in_box = list(filter(lambda e: e.location in subbox, entities))
    #         not_in_box = filter(lambda e: e.location not in subbox, entities)
    #
    #         in_box_copy = deepcopy(in_box)
    #
    #         entity_map[chunk.coordinates] = [
    #             not_in_box,
    #             in_box,
    #         ]  # First index is the list of entities not in the box, the second is for ones that are
    #
    #         yield chunk.coordinates, in_box_copy
    #
    #         if (
    #             in_box != in_box_copy
    #         ):  # If an entity has been changed, update the dictionary entry
    #             entity_map[chunk.coordinates][1] = in_box_copy
    #         else:  # Delete the entry otherwise
    #             del entity_map[chunk.coordinates]
    #
    #     for chunk_coords, entity_list_list in entity_map.items():
    #         chunk = self.get_chunk(*chunk_coords)
    #         in_place_entities = list(
    #             filter(
    #                 lambda e: chunk_coords
    #                 == entity_position_to_chunk_coordinates(e.location),
    #                 entity_list_list[1],
    #             )
    #         )
    #         out_of_place = filter(
    #             lambda e: chunk_coords
    #             != entity_position_to_chunk_coordinates(e.location),
    #             entity_list_list[1],
    #         )
    #
    #         chunk.entities = in_place_entities + list(entity_list_list[0])
    #
    #         if out_of_place:
    #             out_of_place_entities.extend(out_of_place)
    #
    #     if out_of_place_entities:
    #         self.add_entities(out_of_place_entities)
    #
    # def add_entities(self, entities):
    #     proper_entity_chunks = map(
    #         lambda e: (entity_position_to_chunk_coordinates(e.location), e,), entities,
    #     )
    #     accumulated_entities: Dict[Tuple[int, int], List[object]] = {}
    #
    #     for chunk_coord, ent in proper_entity_chunks:
    #         if chunk_coord in accumulated_entities:
    #             accumulated_entities[chunk_coord].append(ent)
    #         else:
    #             accumulated_entities[chunk_coord] = [ent]
    #
    #     for chunk_coord, ents in accumulated_entities.items():
    #         chunk = self.get_chunk(*chunk_coord)
    #
    #         chunk.entities += ents
    #
    # def delete_entities(self, entities):
    #     chunk_entity_pairs = map(
    #         lambda e: (entity_position_to_chunk_coordinates(e.location), e,), entities,
    #     )
    #
    #     for chunk_coord, ent in chunk_entity_pairs:
    #         chunk = self.get_chunk(*chunk_coord)
    #         entities = chunk.entities
    #         entities.remove(ent)
    #         chunk.entities = entities

    def run_operation(
        self, operation: OperationType, dimension: Dimension, *args, create_undo=True
    ) -> Any:
        try:
            out = operation(self, dimension, *args)
            if isinstance(out, GeneratorType):
                try:
                    while True:
                        next(out)
                except StopIteration as e:
                    out = e.value
        except Exception as e:
            self.restore_last_undo_point()
            raise e
        if create_undo:
            self.create_undo_point()
        return out

    def undo(self):
        """Undoes the last set of changes to the world"""
        self._history_manager.undo()

    def redo(self):
        """Redoes the last set of changes to the world"""
        self._history_manager.redo()

    def restore_last_undo_point(self):
        """Restore the world to the state it was when self.create_undo_point was called.
        If an operation errors there may be modifications made that did not get tracked.
        This will revert those changes."""
        self._history_manager.restore_last_undo_point()



from __future__ import annotations

from amulet.world_interface.chunk.interfaces.leveldb.leveldb_6.interface import (
    LevelDB6Interface,
)


class LevelDB7Interface(LevelDB6Interface):
    def __init__(self):
        super().__init__()

        self.features["chunk_version"] = 7


INTERFACE_CLASS = LevelDB7Interface



from __future__ import annotations

from typing import TYPE_CHECKING

from amulet.api.selection import SelectionGroup
from amulet.api.block import Block
from amulet.api.data_types import Dimension, OperationReturnType

if TYPE_CHECKING:
    from amulet.api.world import World


def fill(
    world: "World", dimension: Dimension, target_box: SelectionGroup, fill_block: Block
) -> OperationReturnType:
    if not isinstance(fill_block, Block):
        raise Exception("Fill operation was not given a Block object")
    internal_id = world.palette.get_add_block(fill_block)

    iter_count = len(list(world.get_chunk_slices(target_box, dimension, True)))
    count = 0

    for chunk, slices, _ in world.get_chunk_slices(target_box, dimension, True):
        chunk.blocks[slices] = internal_id
        chunk.changed = True
        count += 1
        yield count / iter_count



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from amulet import log
from amulet.world_interface import formats
from amulet.api.world import World
from amulet.api.errors import FormatLoaderInvalidFormat, FormatLoaderMismatched

if TYPE_CHECKING:
    from amulet.api.wrapper.world_format_wrapper import WorldFormatWrapper


def load_world(directory: str, _format: str = None, forced: bool = False) -> World:
    """
    Creates a Format loader from the given inputs and wraps it in a World class
    :param directory:
    :param _format:
    :param forced:
    :return:
    """
    log.info(f"Loading world {directory}")
    return World(directory, load_format(directory, _format, forced))


def load_format(
    directory: str, _format: str = None, forced: bool = False
) -> "WorldFormatWrapper":
    """
    Loads the world located at the given directory with the appropriate format loader.

    :param directory: The directory of the world
    :param _format: The format name to use
    :param forced: Whether to force load the world even if incompatible
    :return: The loaded world
    """
    if not os.path.isdir(directory):
        if os.path.exists(directory):
            log.error(f'The path "{directory}" does exist but it is not a directory')
            raise Exception(
                f'The path "{directory}" does exist but it is not a directory'
            )
        else:
            log.error(
                f'The path "{directory}" is not a valid path on this file system.'
            )
            raise Exception(
                f'The path "{directory}" is not a valid path on this file system.'
            )
    if _format is not None:
        if _format not in formats.loader:
            log.error(f"Could not find _format loader {_format}")
            raise FormatLoaderInvalidFormat(f"Could not find _format loader {_format}")
        if not forced and not formats.loader.identify(directory) == _format:
            log.error(f"{_format} is incompatible")
            raise FormatLoaderMismatched(f"{_format} is incompatible")
        format_class = formats.loader.get_by_id(_format)
    else:
        format_class = formats.loader.get(directory)
    return format_class(directory)


if __name__ == "__main__":
    import sys
    from amulet.api.block import Block
    import numpy

    cx, cz = 0, 0

    w = load_world(sys.argv[1])
    c = w.get_chunk(cx, cz, "overworld")
    for block in c.blocks[0, :, 0].ravel():  # the blockstates of one vertical column
        print(w.palette[block])
    air = w.palette.get_add_block(
        Block(namespace="universal_minecraft", base_name="air")
    )
    # blocks[0, 30, 0] = stone
    blocks = numpy.random.randint(0, len(w.palette.blocks()), size=(16, 256, 16))
    for index, block in enumerate(w.palette.blocks()):
        if block.base_name in ["lava", "water"]:
            blocks[blocks == index] = air
    c.blocks[:, :, :] = blocks
    w.save()
    w.close()

    w = load_world(sys.argv[1])
    c = w.get_chunk(cx, cz, "overworld")
    for block in c.blocks[0, :, 0].ravel():  # the blockstates of one vertical column
        print(w.palette[block])



VERSION_NUMBER = (0, 0, 0)
VERSION_INT = -1
VERSION_STAGE = "DEV"

if __debug__:
    VERSION_STRING = (
        f"{'.'.join((str(n) for n in VERSION_NUMBER))}-{VERSION_STAGE}-source"
    )
else:
    VERSION_STRING = f"{'.'.join((str(n) for n in VERSION_NUMBER))}-{VERSION_STAGE}"



from __future__ import annotations

from typing import Union, Generator, Dict, Optional, Tuple

from .block import Block
from .chunk import Chunk
from .selection import SelectionBox
from .data_types import DimensionCoordinates

ChunkCache = Dict[DimensionCoordinates, Optional[Chunk]]


class BaseStructure:
    @property
    def sub_chunk_size(self) -> int:
        return 16

    @property
    def chunk_size(self) -> Tuple[int, int, int]:
        return self.sub_chunk_size, self.sub_chunk_size * 16, self.sub_chunk_size

    def get_chunk(self, *args, **kwargs) -> Chunk:
        raise NotImplementedError

    def get_block(self, *args, **kwargs) -> Block:
        raise NotImplementedError

    def _chunk_box(
        self,
        cx: int,
        cz: int,
        chunk_size: Optional[Tuple[int, Union[int, None], int]] = None,
    ):
        """Get a SelectionBox containing the whole of a given chunk"""
        if chunk_size is None:
            chunk_size = self.chunk_size
        return SelectionBox.create_chunk_box(cx, cz, chunk_size[0])

    def get_chunk_boxes(
        self, *args, **kwargs
    ) -> Generator[Tuple[Chunk, SelectionBox], None, None]:
        raise NotImplementedError

    def get_chunk_slices(
        self, *args, **kwargs
    ) -> Generator[Tuple[Chunk, Tuple[slice, slice, slice], SelectionBox], None, None]:
        raise NotImplementedError



import os

from .api import *
from .utils.log import log

entity_support = False
IMG_DIRECTORY = os.path.join(os.path.dirname(__file__), "img")

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



from __future__ import annotations

import numpy
from typing import Tuple, Any, Union, TYPE_CHECKING, Optional

from amulet.api.block_entity import BlockEntity
from amulet.api.entity import Entity
from amulet.api.data_types import AnyNDArray
import amulet_nbt

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator
    from amulet.api.chunk import Chunk


class Interface:
    def decode(self, cx: int, cz: int, data: Any) -> Tuple["Chunk", AnyNDArray]:
        """
        Create an amulet.api.chunk.Chunk object from raw data given by the format
        :param cx: chunk x coordinate
        :type cx: int
        :param cz: chunk z coordinate
        :type cz: int
        :param data: Raw chunk data provided by the format.
        :type data: Any
        :return: Chunk object in version-specific format, along with the block_palette for that chunk.
        :rtype: Chunk
        """
        raise NotImplementedError()

    def _decode_entity(
        self, nbt: amulet_nbt.NBTFile, id_type: str, coord_type: str
    ) -> Optional[Entity]:
        entity = self._decode_base_entity(nbt, id_type, coord_type)
        if entity is not None:
            namespace, base_name, x, y, z, nbt = entity
            return Entity(
                namespace=namespace, base_name=base_name, x=x, y=y, z=z, nbt=nbt
            )

    def _decode_block_entity(
        self, nbt: amulet_nbt.NBTFile, id_type: str, coord_type: str
    ) -> Optional[BlockEntity]:
        entity = self._decode_base_entity(nbt, id_type, coord_type)
        if entity is not None:
            namespace, base_name, x, y, z, nbt = entity
            return BlockEntity(
                namespace=namespace, base_name=base_name, x=x, y=y, z=z, nbt=nbt
            )

    @staticmethod
    def _decode_base_entity(
        nbt: amulet_nbt.NBTFile, id_type: str, coord_type: str
    ) -> Optional[
        Tuple[
            str,
            str,
            Union[int, float],
            Union[int, float],
            Union[int, float],
            amulet_nbt.NBTFile,
        ]
    ]:
        if not isinstance(nbt, amulet_nbt.NBTFile) and isinstance(
            nbt.value, amulet_nbt.TAG_Compound
        ):
            return

        if id_type == "namespace-str-id":
            entity_id = nbt.pop("id", amulet_nbt.TAG_String(""))
            if (
                not isinstance(entity_id, amulet_nbt.TAG_String)
                or entity_id.value == ""
                or ":" not in entity_id.value
            ):
                return
            namespace, base_name = entity_id.value.split(":", 1)

        elif id_type == "str-id":
            entity_id = nbt.pop("id", amulet_nbt.TAG_String(""))
            if (
                not isinstance(entity_id, amulet_nbt.TAG_String)
                or entity_id.value == ""
            ):
                return
            namespace = ""
            base_name = entity_id.value

        elif id_type in ["namespace-str-identifier", "int-id"]:
            if "identifier" in nbt:
                entity_id = nbt.pop("identifier")
                if (
                    not isinstance(entity_id, amulet_nbt.TAG_String)
                    or entity_id.value == ""
                    or ":" not in entity_id.value
                ):
                    return
                namespace, base_name = entity_id.value.split(":", 1)
            elif "id" in nbt:
                entity_id = nbt.pop("id")
                if not isinstance(entity_id, amulet_nbt.TAG_Int):
                    return
                namespace = ""
                base_name = str(entity_id.value)
            else:
                return
        else:
            raise NotImplementedError(f"Entity id type {id_type}")

        if coord_type in ["Pos-list-double", "Pos-list-float"]:
            if "Pos" not in nbt:
                return
            pos = nbt.pop("Pos")
            pos: amulet_nbt.TAG_List
            if not (5 <= pos.list_data_type <= 6 and len(pos) == 3):
                return
            x, y, z = [c.value for c in pos]
        elif coord_type == "xyz-int":
            if not all(
                c in nbt and isinstance(nbt[c], amulet_nbt.TAG_Int)
                for c in ("x", "y", "z")
            ):
                return
            x, y, z = [nbt.pop(c).value for c in ("x", "y", "z")]
        else:
            raise NotImplementedError(f"Entity coord type {coord_type}")

        return namespace, base_name, x, y, z, nbt

    def encode(
        self,
        chunk: "Chunk",
        palette: AnyNDArray,
        max_world_version: Tuple[str, Union[int, Tuple[int, int, int]]],
    ) -> Any:
        """
        Take a version-specific chunk and encode it to raw data for the format to store.
        :param chunk: The already translated version-specfic chunk to encode.
        :type chunk: Chunk
        :param palette: The block_palette the ids in the chunk correspond to.
        :type palette: numpy.ndarray[Block]
        :param max_world_version: The key to use to find the encoder.
        :type max_world_version: Tuple[str, Union[int, Tuple[int, int, int]]]
        :return: Raw data to be stored by the Format.
        :rtype: Any
        """
        raise NotImplementedError()

    def _encode_entity(
        self, entity: Entity, id_type: str, coord_type: str
    ) -> Optional[amulet_nbt.NBTFile]:
        return self._encode_base_entity(entity, id_type, coord_type)

    def _encode_block_entity(
        self, entity: BlockEntity, id_type: str, coord_type: str
    ) -> Optional[amulet_nbt.NBTFile]:
        return self._encode_base_entity(entity, id_type, coord_type)

    @staticmethod
    def _encode_base_entity(
        entity: Union[Entity, BlockEntity], id_type: str, coord_type: str
    ) -> Optional[amulet_nbt.NBTFile]:
        if not isinstance(entity.nbt, amulet_nbt.NBTFile) and isinstance(
            entity.nbt.value, amulet_nbt.TAG_Compound
        ):
            return
        nbt = entity.nbt

        if id_type == "namespace-str-id":
            nbt["id"] = amulet_nbt.TAG_String(entity.namespaced_name)
        elif id_type == "namespace-str-identifier":
            nbt["identifier"] = amulet_nbt.TAG_String(entity.namespaced_name)
        elif id_type == "str-id":
            nbt["id"] = amulet_nbt.TAG_String(entity.base_name)
        elif id_type == "int-id":
            if not entity.base_name.isnumeric():
                return
            nbt["id"] = amulet_nbt.TAG_Int(int(entity.base_name))
        else:
            raise NotImplementedError(f"Entity id type {id_type}")

        if coord_type == "Pos-list-double":
            nbt["Pos"] = amulet_nbt.TAG_List(
                [
                    amulet_nbt.TAG_Double(float(entity.x)),
                    amulet_nbt.TAG_Double(float(entity.y)),
                    amulet_nbt.TAG_Double(float(entity.z)),
                ]
            )
        elif coord_type == "Pos-list-float":
            nbt["Pos"] = amulet_nbt.TAG_List(
                [
                    amulet_nbt.TAG_Float(float(entity.x)),
                    amulet_nbt.TAG_Float(float(entity.y)),
                    amulet_nbt.TAG_Float(float(entity.z)),
                ]
            )
        elif coord_type == "xyz-int":
            nbt["x"] = amulet_nbt.TAG_Int(int(entity.x))
            nbt["y"] = amulet_nbt.TAG_Int(int(entity.y))
            nbt["z"] = amulet_nbt.TAG_Int(int(entity.z))
        else:
            raise NotImplementedError(f"Entity coord type {coord_type}")

        return nbt

    def get_translator(
        self,
        max_world_version: Tuple[str, Union[int, Tuple[int, int, int]]],
        data: Any = None,
    ) -> Tuple["Translator", Union[int, Tuple[int, int, int]]]:
        """
        Get the Translator class for the requested version.
        :param max_world_version: The game version the world was last opened in.
        :type max_world_version: Java: int (DataVersion)    Bedrock: Tuple[int, int, int, ...] (game version number)
        :param data: Optional data to get translator based on chunk version rather than world version
        :param data: Any
        :return: Tuple[Translator, version number for PyMCTranslate to use]
        :rtype: Tuple[Translator, Union[int, Tuple[int, int, int]]]
        """
        raise NotImplementedError

    @staticmethod
    def is_valid(key: Tuple) -> bool:
        """
        Returns whether this Interface is able to interface with the chunk type with a given identifier key,
        generated by the format.

        :param key: The key who's decodability needs to be checked.
        :return: True if this interface can interface with the chunk version associated with the key, False otherwise.
        """
        raise NotImplementedError()



import os
from typing import Optional, Union, Tuple, Generator, TYPE_CHECKING

from amulet import log
from amulet.api.data_types import (
    VersionNumberAny,
    PathOrBuffer,
    ChunkCoordinates,
    AnyNDArray,
)
from amulet.api.wrapper import StructureFormatWrapper
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.errors import ObjectReadError, ObjectWriteError

from .mcstructure import MCStructureWriter, MCStructureReader, MCStructureChunk
from .interface import MCStructureInterface

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator, Interface

mcstructure_interface = MCStructureInterface()


class MCStructureFormatWrapper(StructureFormatWrapper):
    def __init__(self, path: PathOrBuffer, mode: str = "r"):
        super().__init__(path)
        assert mode in ("r", "w"), 'Mode must be either "r" or "w".'
        self._mode = mode
        if isinstance(path, str):
            assert path.endswith(".mcstructure"), 'Path must end with ".mcstructure"'
            if mode == "r":
                assert os.path.isfile(path), "File specified does not exist."
        self._data: Optional[Union[MCStructureWriter, MCStructureReader]] = None
        self._open = False
        self._platform = "bedrock"
        self._version = (1, 15, 2)
        self._selection: SelectionBox = SelectionBox((0, 0, 0), (0, 0, 0))

    def open(self):
        """Open the database for reading and writing"""
        if self._open:
            return
        if self._mode == "r":
            assert (
                isinstance(self.path_or_buffer, str)
                and os.path.isfile(self.path_or_buffer)
            ) or hasattr(self.path_or_buffer, "read"), "File specified does not exist."
            self._data = MCStructureReader(self.path_or_buffer)
            self._selection = self._data.selection
        else:
            self._data = MCStructureWriter(self.path_or_buffer, self._selection,)
        self._open = True

    @property
    def readable(self) -> bool:
        """Can this object have data read from it."""
        return self._mode == "r"

    @property
    def writeable(self) -> bool:
        """Can this object have data written to it."""
        return self._mode == "w"

    @staticmethod
    def is_valid(path: PathOrBuffer) -> bool:
        """
        Returns whether this format is able to load the given object.

        :param path: The path of the object to load.
        :return: True if the world can be loaded by this format, False otherwise.
        """
        return os.path.isfile(path) and path.endswith(".mcstructure")

    @property
    def platform(self) -> str:
        """Platform string ("bedrock" / "java" / ...)"""
        return self._platform

    @property
    def version(self) -> Tuple[int, int, int]:
        return self._version

    @version.setter
    def version(self, version: Tuple[int, int, int]):
        if self._open:
            log.error(
                "mcstructure version cannot be changed after the object has been opened."
            )
            return
        self._version = version

    @property
    def selection(self) -> SelectionGroup:
        """The box that is selected"""
        return SelectionGroup([self._selection])

    @selection.setter
    def selection(self, selection: SelectionGroup):
        if self._open:
            log.error(
                "mcstructure selection cannot be changed after the object has been opened."
            )
            return
        if selection.selection_boxes:
            self._selection = selection.selection_boxes[0]
        else:
            raise Exception("Given selection box is empty")

    def _get_interface(
        self, max_world_version, raw_chunk_data=None
    ) -> "MCStructureInterface":
        return mcstructure_interface

    def _get_interface_and_translator(
        self, max_world_version, raw_chunk_data=None
    ) -> Tuple["Interface", "Translator", VersionNumberAny]:
        interface = self._get_interface(max_world_version, raw_chunk_data)
        translator, version_identifier = interface.get_translator(
            max_world_version, raw_chunk_data, self.translation_manager
        )
        return interface, translator, version_identifier

    @property
    def has_lock(self) -> bool:
        """Verify that the world database can be read and written"""
        return True

    def save(self):
        """Save the data back to the disk database"""
        pass

    def close(self):
        """Close the disk database"""
        self._data.close()

    def unload(self):
        """Unload data stored in the Format class"""
        pass

    def all_chunk_coords(self, *args) -> Generator[ChunkCoordinates, None, None]:
        """A generator of all chunk coords"""
        if self._mode == "r":
            yield from self._data.chunk_coords
        else:
            raise ObjectReadError("all_chunk_coords is only valid in read mode")

    def _encode(
        self, chunk: Chunk, chunk_palette: AnyNDArray, interface: MCStructureInterface,
    ):
        return interface.encode(
            chunk,
            chunk_palette,
            self.max_world_version,
            SelectionBox.create_chunk_box(chunk.cx, chunk.cz).intersection(
                self._selection
            ),
        )

    def delete_chunk(self, cx: int, cz: int, *args):
        raise ObjectWriteError(
            "delete_chunk is not a valid method for an mcstructure file"
        )

    def _put_raw_chunk_data(self, cx: int, cz: int, section: MCStructureChunk, *args):
        """
        Actually stores the data from the interface to disk.
        """
        if self._mode == "w":
            self._data.write(section)
        else:
            raise ObjectWriteError("The mcstructure file is not open for writing.")

    def _get_raw_chunk_data(self, cx: int, cz: int, *args) -> MCStructureChunk:
        """
        Return the interface key and data to interface with given chunk coordinates.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :return: The raw chunk data.
        """
        if self._mode == "r":
            return self._data.read(cx, cz)
        else:
            raise ObjectReadError("The mcstructure file is not open for reading.")



from amulet.api.selection import SelectionGroup
from amulet.api.structure import Structure
from amulet.api.data_types import Dimension
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from amulet.api.world import World


def clone(world: "World", dimension: Dimension, source: SelectionGroup, target: dict):
    dst_location = (target.get("x", 0), target.get("y", 0), target.get("z", 0))
    structure = Structure.from_world(world, source, dimension)
    for (
        src_chunk,
        src_slices,
        _,
        (dst_cx, dst_cz),
        dst_slices,
        _,
    ) in structure.get_moved_chunk_slices(dst_location):
        dst_chunk = world.get_chunk(dst_cx, dst_cz, dimension)
        dst_chunk.blocks[dst_slices] = src_chunk.blocks[src_slices]
        dst_chunk.changed = True



from .format_wrapper import BaseFormatWrapper
from amulet.api.data_types import PathOrBuffer


class StructureFormatWrapper(BaseFormatWrapper):
    def __init__(self, path_or_buffer: PathOrBuffer):
        super().__init__()
        self._path_or_buffer = path_or_buffer

    @property
    def path_or_buffer(self) -> PathOrBuffer:
        return self._path_or_buffer
